/* MIT License
 *
 * Copyright (c) 2016-2020 INRIA, CMU and Microsoft Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */


#ifndef __Hacl_HMAC_H
#define __Hacl_HMAC_H

#if defined(__cplusplus)
extern "C" {
#endif

#include "evercrypt_targetconfig.h"
#include "libintvector.h"
#include "kremlin/internal/types.h"
#include "kremlin/lowstar_endianness.h"
#include <string.h>
#include "kremlin/internal/target.h"


#include "Hacl_Kremlib.h"
#include "Hacl_Impl_Blake2_Constants.h"
#include "Hacl_Hash.h"

/* SNIPPET_START: Hacl_HMAC_legacy_compute_sha1 */

void
Hacl_HMAC_legacy_compute_sha1(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_legacy_compute_sha1 */

/* SNIPPET_START: Hacl_HMAC_compute_sha2_256 */

void
Hacl_HMAC_compute_sha2_256(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_compute_sha2_256 */

/* SNIPPET_START: Hacl_HMAC_compute_sha2_384 */

void
Hacl_HMAC_compute_sha2_384(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_compute_sha2_384 */

/* SNIPPET_START: Hacl_HMAC_compute_sha2_512 */

void
Hacl_HMAC_compute_sha2_512(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_compute_sha2_512 */

/* SNIPPET_START: K____uint32_t__uint64_t */

typedef struct K____uint32_t__uint64_t_s
{
  uint32_t *fst;
  uint64_t snd;
}
K____uint32_t__uint64_t;

/* SNIPPET_END: K____uint32_t__uint64_t */

/* SNIPPET_START: Hacl_HMAC_compute_blake2s_32 */

void
Hacl_HMAC_compute_blake2s_32(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_compute_blake2s_32 */

/* SNIPPET_START: K____uint64_t__FStar_UInt128_uint128 */

typedef struct K____uint64_t__FStar_UInt128_uint128_s
{
  uint64_t *fst;
  FStar_UInt128_uint128 snd;
}
K____uint64_t__FStar_UInt128_uint128;

/* SNIPPET_END: K____uint64_t__FStar_UInt128_uint128 */

/* SNIPPET_START: Hacl_HMAC_compute_blake2b_32 */

void
Hacl_HMAC_compute_blake2b_32(
  uint8_t *dst,
  uint8_t *key,
  uint32_t key_len,
  uint8_t *data,
  uint32_t data_len
);

/* SNIPPET_END: Hacl_HMAC_compute_blake2b_32 */

#if defined(__cplusplus)
}
#endif

#define __Hacl_HMAC_H_DEFINED
#endif
